#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""The geopm python package: launcher, error, endpoint, io, pio, plotter,
policy_store, topo, agent, and version.

"""

from .version import version as __version__
